#include "AdvancedOgreFramework.h"

#include "OGRE/Overlay/OgreOverlay.h"

using namespace Ogre;

template<> AdvancedOgreFramework* Ogre::Singleton<AdvancedOgreFramework>::msSingleton = 0;

AdvancedOgreFramework::AdvancedOgreFramework()
{
	m_pRoot = 0;
	m_pRenderWnd = 0;
	m_pViewport = 0;
	m_pLog = 0;
	m_pTimer = 0;

	m_pInputMgr = 0;
	m_pKeyboard = 0;
	m_pMouse = 0;
	m_pTrayMgr = 0;
}

AdvancedOgreFramework::~AdvancedOgreFramework()
{
	AdvancedOgreFramework::getSingletonPtr()->m_pLog->logMessage("Shutdown OGRE...");

	if(m_pTrayMgr)
	{
		delete m_pTrayMgr;
		m_pTrayMgr = 0;
	}

	if(mOverlaySystem)
	{
		delete mOverlaySystem;
		mOverlaySystem = 0;
	}

	if(m_pInputMgr)
	{
		OIS::InputManager::destroyInputSystem(m_pInputMgr);
	}

	if(m_pRoot)
	{
		delete m_pRoot;
		m_pRoot = 0;
	}

}

bool AdvancedOgreFramework::initOgre(Ogre::String wndTitle, OIS::KeyListener *pKeyListener, OIS::MouseListener *pMouseListener)
{
	Ogre::LogManager* logMgr = new Ogre::LogManager();

	m_pLog = Ogre::LogManager::getSingleton().createLog("OgreLogfile.log", true, true, false);
	m_pLog->setDebugOutputEnabled(true);

	m_pRoot = new Ogre::Root();

	if(!m_pRoot->showConfigDialog())
		return false;
	m_pRenderWnd = m_pRoot->initialise(true, wndTitle);

	m_pViewport = m_pRenderWnd->addViewport(0);
	m_pViewport->setBackgroundColour(ColourValue(0.5f, 0.5f, 0.5f, 1.0f));

	m_pViewport->setCamera(0);

	size_t hWnd = 0;
	OIS::ParamList paramList;
	m_pRenderWnd->getCustomAttribute("WINDOW", &hWnd);


	// Don't grab the input
	paramList.insert(OIS::ParamList::value_type("WINDOW", Ogre::StringConverter::toString(hWnd)));
	paramList.insert(std::make_pair(std::string("x11_mouse_grab"), std::string("false")));
	paramList.insert(std::make_pair(std::string("x11_mouse_hide"), std::string("false")));
	paramList.insert(std::make_pair(std::string("x11_keyboard_grab"), std::string("true")));
	paramList.insert(std::make_pair(std::string("XAutoRepeatOn"), std::string("true")));
	
	
	m_pInputMgr = OIS::InputManager::createInputSystem(paramList);

	m_pKeyboard = static_cast<OIS::Keyboard*>(m_pInputMgr->createInputObject(OIS::OISKeyboard, true));
	m_pMouse = static_cast<OIS::Mouse*>(m_pInputMgr->createInputObject(OIS::OISMouse, true));

	m_pMouse->getMouseState().height = m_pRenderWnd->getHeight();
	m_pMouse->getMouseState().width  = m_pRenderWnd->getWidth();
	
	

	if(pKeyListener == 0)
		m_pKeyboard->setEventCallback(this);
	else
		m_pKeyboard->setEventCallback(pKeyListener);

	if(pMouseListener == 0)
		m_pMouse->setEventCallback(this);
	else
		m_pMouse->setEventCallback(pMouseListener);

	Ogre::String secName, typeName, archName;
	Ogre::ConfigFile cf;
	cf.load("resources.cfg");
	mOverlaySystem = new Ogre::OverlaySystem();
	
	Ogre::ConfigFile::SectionIterator seci = cf.getSectionIterator();
	while (seci.hasMoreElements())
	{
		secName = seci.peekNextKey();
		Ogre::ConfigFile::SettingsMultiMap *settings = seci.getNext();
		Ogre::ConfigFile::SettingsMultiMap::iterator i;
		for (i = settings->begin(); i != settings->end(); ++i)
		{
			typeName = i->first;
			archName = i->second;
			Ogre::ResourceGroupManager::getSingleton().addResourceLocation(archName, typeName, secName);
		}
	}
	Ogre::TextureManager::getSingleton().setDefaultNumMipmaps(5);
	Ogre::ResourceGroupManager::getSingleton().initialiseAllResourceGroups();

	OgreBites::InputContext inputContext;
	inputContext.mMouse = m_pMouse;
	inputContext.mKeyboard = m_pKeyboard;

	m_pTrayMgr = new OgreBites::SdkTrayManager("AOFTrayMgr", m_pRenderWnd, inputContext, 0);


	m_pTimer = new Ogre::Timer();
	m_pTimer->reset();

	m_pRenderWnd->setActive(true);

	return true;
}

bool AdvancedOgreFramework::keyPressed(const OIS::KeyEvent &keyEventRef)
{
	if(m_pKeyboard->isKeyDown(OIS::KC_SYSRQ))
	{
		m_pRenderWnd->writeContentsToTimestampedFile("AOF_Screenshot_", ".jpg");
		return true;
	}

	if(m_pKeyboard->isKeyDown(OIS::KC_O))
	{
		if(m_pTrayMgr->isLogoVisible())
		{
			m_pTrayMgr->hideFrameStats();
			m_pTrayMgr->hideLogo();
		}
		else
		{
			m_pTrayMgr->showFrameStats(OgreBites::TL_BOTTOMLEFT);
			m_pTrayMgr->showLogo(OgreBites::TL_BOTTOMRIGHT);
		}
	}

	return true;
}

bool AdvancedOgreFramework::keyReleased(const OIS::KeyEvent &keyEventRef)
{
	return true;
}

bool AdvancedOgreFramework::mouseMoved(const OIS::MouseEvent &evt)
{
	return true;
}

bool AdvancedOgreFramework::mousePressed(const OIS::MouseEvent &evt, OIS::MouseButtonID id)
{
	return true;
}

bool AdvancedOgreFramework::mouseReleased(const OIS::MouseEvent &evt, OIS::MouseButtonID id)
{
	return true;
}

double AdvancedOgreFramework::getCurrentTime()
{

	static struct timeval start;

	gettimeofday(&start, NULL);
	return ((start.tv_sec * 1000) + (start.tv_usec / 1000.0));

}

void AdvancedOgreFramework::updateOgre(double timeSinceLastFrame)
{
}
