#include "VehicleController.h"
#include "AdvancedOgreFramework.h"

#include <OgreSceneNode.h>
#include <OgreSceneManager.h>
#include <btBulletCollisionCommon.h>
#include <BtOgrePG.h>

VehicleController::VehicleController(btDynamicsWorld *phyWorld, SceneManager *sceneMgr) : mPhyWorld(phyWorld) , mSceneMgr(sceneMgr)
{
	gEngineForce = 1000.f;
	gBreakingForce = 0.f;
	gVehicleSteering = 0.f;
	
	mEngineForceForward = 2000 * 6;
	mEngineForceBack = -2000 * 4;
	mBrakeForce = 80;
}

VehicleController::~VehicleController()
{
	delete m_vehicle;
	delete m_vehicleRayCaster;
	delete m_carChassis->getMotionState();
	delete m_carChassis;
	delete chassisShape;
	
}

void VehicleController::setUp()
{
#define CAR_WIDTH 7.5
	
	btScalar suspensionRestLength(5.5);

	wheelRadius = 3.2f;
	wheelFriction = 10000;//BT_LARGE_FLOAT;
	suspensionStiffness = 4.f;
	suspensionDamping = 5.3f;
	suspensionCompression = 10.4f;
	rollInfluence = 0.0f;
	int rightIndex = 0;
	int upIndex = 1;
	int forwardIndex = 2;
	btVector3 wheelDirectionCS0(0,-1,0);
	btVector3 wheelAxleCS(-1,0,0);
	
	
	//! Car Body
	
	mCarBody = mSceneMgr->getRootSceneNode()->createChildSceneNode("Car");
	Ogre::Entity* carEntity = mSceneMgr->createEntity("Dodge", "CarBody.mesh");
	carEntity->setCastShadows(true);
	Ogre::SceneNode* carNode = mCarBody->createChildSceneNode("Dodge");
	carNode->attachObject(carEntity);
	chassisShape = new btBoxShape(btVector3(10, 4, 25));
	
	m_carChassis = localCreateRigidBody(mCarBody, 800, chassisShape);//chassisShape);
	m_carChassis->setDamping(0.2, 0.2);
	
	//	btCylinderShapeX *m_wheelShape = new btCylinderShapeX(btVector3(wheelWidth,wheelRadius,wheelRadius));
	
	
	btRaycastVehicle::btVehicleTuning	m_tuning;
	m_tuning.m_frictionSlip = 200;
	/// create vehicle
	{
		
		m_vehicleRayCaster = new btDefaultVehicleRaycaster(mPhyWorld);
		m_vehicle = new btRaycastVehicle(m_tuning,m_carChassis,m_vehicleRayCaster);
		
		///never deactivate the vehicle
		m_carChassis->setActivationState(DISABLE_DEACTIVATION);
		
		mPhyWorld->addVehicle(m_vehicle);
		
		float connectionHeight = 1.2;
		
		
		bool isFrontWheel=true;
		
		//choose coordinate system
		m_vehicle->setCoordinateSystem(rightIndex,upIndex,forwardIndex);
		
		float frontOffset = 16;
		float backOffset = -13;
		
		// Front wheels
		btVector3 connectionPointCS0(CAR_WIDTH,connectionHeight,frontOffset);
		m_vehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,wheelRadius,m_tuning,isFrontWheel);
		
		connectionPointCS0 = btVector3(-CAR_WIDTH,connectionHeight,frontOffset);
		m_vehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,wheelRadius,m_tuning,isFrontWheel);
		
		
		// Back wheels
		isFrontWheel = false;
		connectionPointCS0 = btVector3(-CAR_WIDTH,connectionHeight,backOffset);
		m_vehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,wheelRadius,m_tuning,isFrontWheel);
		
		connectionPointCS0 = btVector3(CAR_WIDTH,connectionHeight,backOffset);
		m_vehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,wheelRadius,m_tuning,isFrontWheel);
		
		for (int i=0;i<m_vehicle->getNumWheels();i++)
		{
			btWheelInfo& wheel = m_vehicle->getWheelInfo(i);
			wheel.m_suspensionStiffness = suspensionStiffness;
			wheel.m_wheelsDampingRelaxation = suspensionDamping;
			wheel.m_wheelsDampingCompression = suspensionCompression;
			wheel.m_frictionSlip = wheelFriction;
			wheel.m_rollInfluence = rollInfluence;
		}
	}
	
	clientResetScene();
	
	// Tires
	for(int i = 0; i < 4; i++) {
		Ogre::Entity* tireEntity = mSceneMgr->createEntity("Tire.mesh");
		tireEntity->setCastShadows(true);
		tireNode = mSceneMgr->getRootSceneNode()->createChildSceneNode("Tire" + i);
		tireNode->attachObject(tireEntity);
		tireNode->setPosition(Ogre::Vector3(0,0,0));
		tires.push_back(tireNode);
	}
	
}

void VehicleController::updateWheelsParams()
{
	for (int i=2;i<m_vehicle->getNumWheels();i++)
	{
		btWheelInfo& wheel = m_vehicle->getWheelInfo(i);
		wheel.m_suspensionStiffness = suspensionStiffness;
		wheel.m_wheelsDampingRelaxation = suspensionDamping;
		wheel.m_wheelsDampingCompression = suspensionCompression;
		wheel.m_frictionSlip = wheelFriction;
		wheel.m_rollInfluence = rollInfluence;
		wheel.m_suspensionRelativeVelocity = wheelRadius;
	}
}


void VehicleController::clientResetScene()
{
	gVehicleSteering = 0;
	
	btTransform centerOfMass = btTransform::getIdentity();
	centerOfMass.setOrigin(btVector3(0, 10, 0));
	m_carChassis->setCenterOfMassTransform(centerOfMass);
	
	m_carChassis->setLinearVelocity(btVector3(0,0,0));
	m_carChassis->setAngularVelocity(btVector3(0,0,0));
	mPhyWorld->getBroadphase()->getOverlappingPairCache()->cleanProxyFromPairs(m_carChassis->getBroadphaseHandle(), mPhyWorld->getDispatcher());
	if (m_vehicle)
	{
		m_vehicle->resetSuspension();
		for (int i=0;i<m_vehicle->getNumWheels();i++)
		{
			//synchronize the wheels with the (interpolated) chassis worldtransform
			m_vehicle->updateWheelTransform(i,true);
		}
	}
}

bool VehicleController::frameStarted(const FrameEvent &evt)
{
	if(AdvancedOgreFramework::getSingletonPtr()->m_pKeyboard->isKeyDown(OIS::KC_UP))
	{
		m_vehicle->applyEngineForce(mEngineForceForward,0);
		m_vehicle->applyEngineForce(mEngineForceForward,1);
//		m_vehicle->applyEngineForce(mEngineForceForward,2);
//		m_vehicle->applyEngineForce(mEngineForceForward,3);
	}
	else if(AdvancedOgreFramework::getSingletonPtr()->m_pKeyboard->isKeyDown(OIS::KC_DOWN))
	{
		m_vehicle->applyEngineForce(mEngineForceBack,0);
		m_vehicle->applyEngineForce(mEngineForceBack,1);
//		m_vehicle->applyEngineForce(mEngineForceBack,2);
//		m_vehicle->applyEngineForce(mEngineForceBack,3);
	}
	else
	{
		m_vehicle->applyEngineForce(0,0);
		m_vehicle->applyEngineForce(0,1);
//		m_vehicle->applyEngineForce(0,2);
//		m_vehicle->applyEngineForce(0,3);
	}
	
	if(AdvancedOgreFramework::getSingletonPtr()->m_pKeyboard->isKeyDown(OIS::KC_SPACE))
	{
//		m_vehicle->setBrake(80,0);
//		m_vehicle->setBrake(80,1);
		m_vehicle->setBrake(mBrakeForce,2);
		m_vehicle->setBrake(mBrakeForce,3);
		m_vehicle->getWheelInfo(2).m_frictionSlip = 0;
		m_vehicle->getWheelInfo(3).m_frictionSlip = 0;
	}
	else
	{
		m_vehicle->setBrake(0,0);
		m_vehicle->setBrake(0,1);
//		m_vehicle->setBrake(0,2);
//		m_vehicle->setBrake(0,3);
		m_vehicle->getWheelInfo(2).m_frictionSlip = wheelFriction;
		m_vehicle->getWheelInfo(3).m_frictionSlip = wheelFriction;
	}
	
	if(AdvancedOgreFramework::getSingletonPtr()->m_pKeyboard->isKeyDown(OIS::KC_LEFT) && gVehicleSteering < 0.4)
	{
		gVehicleSteering += 0.8 * evt.timeSinceLastFrame;
	}
	
	if(AdvancedOgreFramework::getSingletonPtr()->m_pKeyboard->isKeyDown(OIS::KC_RIGHT) && gVehicleSteering > -0.4)
	{
		gVehicleSteering -= 0.8 * evt.timeSinceLastFrame;
	}
	
	m_vehicle->setSteeringValue(gVehicleSteering,0);
	m_vehicle->setSteeringValue(gVehicleSteering,1);
	
	for (int i=0;i<m_vehicle->getNumWheels();i++)
	{
		//synchronize the wheels with the (interpolated) chassis worldtransform
		m_vehicle->updateWheelTransform(i,true);
		Ogre::Vector3 pos = BtOgre::Convert::toOgre(m_vehicle->getWheelInfo(i).m_worldTransform.getOrigin());
		Ogre::Quaternion orientation = BtOgre::Convert::toOgre(m_vehicle->getWheelInfo(i).m_worldTransform.getRotation());
		
		Ogre::SceneNode *tire = tires[i];
		tire->setPosition(pos);
		tire->setOrientation(orientation);
		if(i == 0 || i == 4 || i == 3)
		{
			tire->rotate(Vector3(0,0,1), Radian(M_PI_2));
		}
		else
		{
			tire->rotate(Vector3(0,0,1), Radian(-M_PI_2));
		}
	}
	
	return true;
	
}

btRigidBody *VehicleController::localCreateRigidBody(SceneNode *node, float mass, btCollisionShape *shape)
{
	
	btAssert((!shape || shape->getShapeType() != INVALID_SHAPE_PROXYTYPE));
	
	//rigidbody is dynamic if and only if mass is non zero, otherwise static
	bool isDynamic = (mass != 0.f);
	
	btVector3 localInertia(0,0,0);
	if (isDynamic)
		shape->calculateLocalInertia(mass,localInertia);
	
	//using motionstate is recommended, it provides interpolation capabilities, and only synchronizes 'active' objects
	
	
	BtOgre::RigidBodyState* myMotionState = new BtOgre::RigidBodyState(node);
	
	btRigidBody::btRigidBodyConstructionInfo cInfo(mass,myMotionState,shape,localInertia);
	
	btRigidBody* body = new btRigidBody(cInfo);
	body->setContactProcessingThreshold(BT_LARGE_FLOAT);
	
	
	mPhyWorld->addRigidBody(body);
	
	return body;
}
